@@ -1,4 +1,4 @@
-// Copyright 2014 The Chromium Authors. All rights reserved.
+// Copyright 2014 The Chromium Authors
 // Use of this source code is governed by a BSD-style license that can be
 // found in the LICENSE file.
 
@@ -9,31 +9,32 @@
 
 #include <memory>
 #include <string>
+#include <string_view>
 #include <utility>
 #include <vector>
 
-#include "base/bind.h"
-#include "base/callback.h"
 #include "base/containers/queue.h"
+#include "base/functional/bind.h"
+#include "base/functional/callback.h"
 #include "base/logging.h"
-#include "base/macros.h"
 #include "base/memory/ref_counted_delete_on_sequence.h"
 #include "base/metrics/histogram_macros.h"
-#include "base/sequenced_task_runner.h"
-#include "base/task/post_task.h"
-#include "base/threading/sequenced_task_runner_handle.h"
+#include "base/strings/string_util.h"
+#include "base/task/sequenced_task_runner.h"
 #include "base/threading/thread_restrictions.h"
-#include "base/threading/thread_task_runner_handle.h"
 #include "chrome/browser/printing/printing_service.h"
 #include "chrome/services/printing/public/mojom/pdf_to_emf_converter.mojom.h"
+#include "chrome/services/printing/public/mojom/printing_service.mojom.h"
+#include "components/device_event_log/device_event_log.h"
 #include "content/public/browser/browser_thread.h"
 #include "content/public/browser/child_process_data.h"
-#include "mojo/public/cpp/bindings/pending_receiver.h"
 #include "mojo/public/cpp/bindings/pending_remote.h"
 #include "mojo/public/cpp/bindings/receiver.h"
 #include "mojo/public/cpp/bindings/remote.h"
 #include "printing/emf_win.h"
 #include "printing/pdf_render_settings.h"
+#include "third_party/abseil-cpp/absl/types/optional.h"
+#include "url/gurl.h"
 
 using content::BrowserThread;
 
@@ -41,67 +42,20 @@ namespace printing {
 
 namespace {
 
-class PdfToEmfConverterClientImpl : public mojom::PdfToEmfConverterClient {
- public:
-  explicit PdfToEmfConverterClientImpl(
-      mojo::PendingReceiver<mojom::PdfToEmfConverterClient> receiver)
-      : receiver_(this, std::move(receiver)) {}
-
- private:
-  // mojom::PdfToEmfConverterClient implementation.
-  void PreCacheFontCharacters(
-      const std::vector<uint8_t>& logfont_data,
-      const base::string16& characters,
-      PreCacheFontCharactersCallback callback) override {
-    // TODO(scottmg): pdf/ppapi still require the renderer to be able to
-    // precache GDI fonts (http://crbug.com/383227), even when using
-    // DirectWrite. Eventually this shouldn't be added and should be moved to
-    // FontCacheDispatcher too. http://crbug.com/356346.
-
-    // First, comments from FontCacheDispatcher::OnPreCacheFont do apply here
-    // too. Except that for True Type fonts, GetTextMetrics will not load the
-    // font in memory. The only way windows seem to load properly, it is to
-    // create a similar device (like the one in which we print), then do an
-    // ExtTextOut, as we do in the printing thread, which is sandboxed.
-    const LOGFONT* logfont =
-        reinterpret_cast<const LOGFONT*>(&logfont_data.at(0));
-
-    HDC hdc = CreateEnhMetaFile(nullptr, nullptr, nullptr, nullptr);
-    HFONT font_handle = CreateFontIndirect(logfont);
-    DCHECK(font_handle != nullptr);
-
-    HGDIOBJ old_font = SelectObject(hdc, font_handle);
-    DCHECK(old_font != nullptr);
-
-    ExtTextOut(hdc, 0, 0, ETO_GLYPH_INDEX, 0, characters.c_str(),
-               characters.length(), nullptr);
-
-    SelectObject(hdc, old_font);
-    DeleteObject(font_handle);
-
-    HENHMETAFILE metafile = CloseEnhMetaFile(hdc);
-
-    if (metafile)
-      DeleteEnhMetaFile(metafile);
-
-    std::move(callback).Run();
-  }
-
-  mojo::Receiver<mojom::PdfToEmfConverterClient> receiver_;
-};
-
 // Emf subclass that knows how to play back PostScript data embedded as EMF
 // comment records.
 class PostScriptMetaFile : public Emf {
  public:
-  PostScriptMetaFile() {}
-  ~PostScriptMetaFile() override {}
+  PostScriptMetaFile() = default;
+
+  PostScriptMetaFile(const PostScriptMetaFile&) = delete;
+  PostScriptMetaFile& operator=(const PostScriptMetaFile&) = delete;
+
+  ~PostScriptMetaFile() override = default;
 
  private:
   // Emf:
   bool SafePlayback(HDC hdc) const override;
-
-  DISALLOW_COPY_AND_ASSIGN(PostScriptMetaFile);
 };
 
 // Class for converting PDF to another format for printing (Emf, Postscript).
@@ -118,7 +72,13 @@ class PdfConverterImpl : public PdfConverter {
  public:
   PdfConverterImpl(scoped_refptr<base::RefCountedMemory> data,
                    const PdfRenderSettings& conversion_settings,
+                   const absl::optional<bool>& use_skia,
+                   const GURL& url,
                    StartCallback start_callback);
+
+  PdfConverterImpl(const PdfConverterImpl&) = delete;
+  PdfConverterImpl& operator=(const PdfConverterImpl&) = delete;
+
   ~PdfConverterImpl() override;
 
   static void set_fail_when_initializing_conversion_for_tests(bool fail) {
@@ -132,34 +92,36 @@ class PdfConverterImpl : public PdfConverter {
  private:
   class GetPageCallbackData {
    public:
-    GetPageCallbackData(int page_number, PdfConverter::GetPageCallback callback)
-        : page_number_(page_number), callback_(callback) {}
+    GetPageCallbackData(uint32_t page_index,
+                        PdfConverter::GetPageCallback callback)
+        : page_index_(page_index), callback_(callback) {}
+
+    GetPageCallbackData(const GetPageCallbackData&) = delete;
+    GetPageCallbackData& operator=(const GetPageCallbackData&) = delete;
 
     GetPageCallbackData(GetPageCallbackData&& other) {
       *this = std::move(other);
     }
 
     GetPageCallbackData& operator=(GetPageCallbackData&& rhs) {
-      page_number_ = rhs.page_number_;
+      page_index_ = rhs.page_index_;
       callback_ = rhs.callback_;
       return *this;
     }
 
-    int page_number() const { return page_number_; }
+    uint32_t page_index() const { return page_index_; }
 
     PdfConverter::GetPageCallback callback() const { return callback_; }
 
    private:
-    int page_number_;
+    uint32_t page_index_;
 
     PdfConverter::GetPageCallback callback_;
-
-    DISALLOW_COPY_AND_ASSIGN(GetPageCallbackData);
   };
 
   void Initialize(scoped_refptr<base::RefCountedMemory> data);
 
-  void GetPage(int page_number,
+  void GetPage(uint32_t page_index,
                PdfConverter::GetPageCallback get_page_callback) override;
 
   void Stop();
@@ -172,11 +134,15 @@ class PdfConverterImpl : public PdfConverter {
   void OnPageDone(base::ReadOnlySharedMemoryRegion emf_region,
                   float scale_factor);
 
-  void OnFailed(const std::string& error_message);
+  void OnFailed(std::string_view error_message);
 
   void RecordConversionMetrics();
 
-  PdfRenderSettings settings_;
+  const PdfRenderSettings settings_;
+
+  absl::optional<bool> use_skia_;
+
+  const GURL url_;
 
   // Document loaded callback.
   PdfConverter::StartCallback start_callback_;
@@ -192,9 +158,6 @@ class PdfConverterImpl : public PdfConverter {
   uint32_t pages_generated_ = 0;
   uint32_t page_count_ = 0;
 
-  std::unique_ptr<PdfToEmfConverterClientImpl>
-      pdf_to_emf_converter_client_impl_;
-
   mojo::Remote<mojom::PdfToEmfConverter> pdf_to_emf_converter_;
 
   mojo::Remote<mojom::PdfToEmfConverterFactory> pdf_to_emf_converter_factory_;
@@ -202,8 +165,6 @@ class PdfConverterImpl : public PdfConverter {
   base::WeakPtrFactory<PdfConverterImpl> weak_ptr_factory_{this};
 
   static bool simulate_failure_initializing_conversion_;
-
-  DISALLOW_COPY_AND_ASSIGN(PdfConverterImpl);
 };
 
 // static
@@ -214,6 +175,8 @@ std::unique_ptr<MetafilePlayer> PdfConverterImpl::GetMetaFileFromMapping(
   std::unique_ptr<Emf> metafile;
   if (settings_.mode == PdfRenderSettings::Mode::POSTSCRIPT_LEVEL2 ||
       settings_.mode == PdfRenderSettings::Mode::POSTSCRIPT_LEVEL3 ||
+      settings_.mode ==
+          PdfRenderSettings::Mode::POSTSCRIPT_LEVEL3_WITH_TYPE42_FONTS ||
       settings_.mode == PdfRenderSettings::Mode::TEXTONLY) {
     metafile = std::make_unique<PostScriptMetaFile>();
   } else {
@@ -243,8 +206,13 @@ bool PostScriptMetaFile::SafePlayback(HDC hdc) const {
 
 PdfConverterImpl::PdfConverterImpl(scoped_refptr<base::RefCountedMemory> data,
                                    const PdfRenderSettings& settings,
+                                   const absl::optional<bool>& use_skia,
+                                   const GURL& url,
                                    StartCallback start_callback)
-    : settings_(settings), start_callback_(std::move(start_callback)) {
+    : settings_(settings),
+      use_skia_(use_skia),
+      url_(url),
+      start_callback_(std::move(start_callback)) {
   DCHECK_CURRENTLY_ON(BrowserThread::UI);
   DCHECK(start_callback_);
 
@@ -259,34 +227,28 @@ PdfConverterImpl::~PdfConverterImpl() {
 
 void PdfConverterImpl::Initialize(scoped_refptr<base::RefCountedMemory> data) {
   if (simulate_failure_initializing_conversion_) {
-    OnFailed(std::string("Failed to create PDF data mapping."));
+    OnFailed("Failed to create PDF data mapping.");
     return;
   }
 
   base::MappedReadOnlyRegion memory =
       base::ReadOnlySharedMemoryRegion::Create(data->size());
   if (!memory.IsValid()) {
-    OnFailed(std::string("Failed to create PDF data mapping."));
+    OnFailed("Failed to create PDF data mapping.");
     return;
   }
 
+  PRINTER_LOG(EVENT) << "PdfConverter created. Mode: " << settings_.mode;
   memcpy(memory.mapping.memory(), data->front(), data->size());
 
   GetPrintingService()->BindPdfToEmfConverterFactory(
       pdf_to_emf_converter_factory_.BindNewPipeAndPassReceiver());
   pdf_to_emf_converter_factory_.set_disconnect_handler(base::BindOnce(
       &PdfConverterImpl::OnFailed, weak_ptr_factory_.GetWeakPtr(),
-      std::string("Connection to PdfToEmfConverterFactory error.")));
-
-  mojo::PendingRemote<mojom::PdfToEmfConverterClient>
-      pdf_to_emf_converter_client_remote;
-  pdf_to_emf_converter_client_impl_ =
-      std::make_unique<PdfToEmfConverterClientImpl>(
-          pdf_to_emf_converter_client_remote.InitWithNewPipeAndPassReceiver());
+      "Connection to PdfToEmfConverterFactory error."));
 
   pdf_to_emf_converter_factory_->CreateConverter(
       std::move(memory.region), settings_,
-      std::move(pdf_to_emf_converter_client_remote),
       base::BindOnce(&PdfConverterImpl::OnPageCount,
                      weak_ptr_factory_.GetWeakPtr()));
 }
@@ -299,26 +261,30 @@ void PdfConverterImpl::OnPageCount(
   pdf_to_emf_converter_.Bind(std::move(converter));
   pdf_to_emf_converter_.set_disconnect_handler(base::BindOnce(
       &PdfConverterImpl::OnFailed, weak_ptr_factory_.GetWeakPtr(),
-      std::string("Connection to PdfToEmfConverter error.")));
+      "Connection to PdfToEmfConverter error."));
+  pdf_to_emf_converter_->SetWebContentsURL(url_);
+  if (use_skia_) {
+    pdf_to_emf_converter_->SetUseSkiaRendererPolicy(*use_skia_);
+  }
   std::move(start_callback_).Run(page_count);
   page_count_ = page_count;
 }
 
 void PdfConverterImpl::GetPage(
-    int page_number,
+    uint32_t page_index,
     PdfConverter::GetPageCallback get_page_callback) {
   DCHECK_CURRENTLY_ON(BrowserThread::UI);
   DCHECK(pdf_to_emf_converter_.is_bound());
 
   // Store callback before any OnFailed() call to make it called on failure.
-  get_page_callbacks_.push(GetPageCallbackData(page_number, get_page_callback));
+  get_page_callbacks_.push(GetPageCallbackData(page_index, get_page_callback));
 
   if (!pdf_to_emf_converter_)
-    return OnFailed(std::string("No PdfToEmfConverter."));
+    return OnFailed("No PdfToEmfConverter.");
 
   pdf_to_emf_converter_->ConvertPage(
-      page_number, base::BindOnce(&PdfConverterImpl::OnPageDone,
-                                  weak_ptr_factory_.GetWeakPtr()));
+      page_index, base::BindOnce(&PdfConverterImpl::OnPageDone,
+                                 weak_ptr_factory_.GetWeakPtr()));
 }
 
 void PdfConverterImpl::OnPageDone(base::ReadOnlySharedMemoryRegion emf_region,
@@ -326,7 +292,7 @@ void PdfConverterImpl::OnPageDone(base::ReadOnlySharedMemoryRegion emf_region,
   DCHECK_CURRENTLY_ON(BrowserThread::UI);
 
   if (get_page_callbacks_.empty())
-    return OnFailed(std::string("No get_page callbacks."));
+    return OnFailed("No get_page callbacks.");
 
   GetPageCallbackData& data = get_page_callbacks_.front();
   std::unique_ptr<MetafilePlayer> metafile;
@@ -343,8 +309,8 @@ void PdfConverterImpl::OnPageDone(base::ReadOnlySharedMemoryRegion emf_region,
   }
 
   base::WeakPtr<PdfConverterImpl> weak_this = weak_ptr_factory_.GetWeakPtr();
-  data.callback().Run(data.page_number(), scale_factor, std::move(metafile));
-  // WARNING: the callback might have deleted |this|!
+  data.callback().Run(data.page_index(), scale_factor, std::move(metafile));
+  // WARNING: the callback might have deleted `this`!
   if (!weak_this)
     return;
   get_page_callbacks_.pop();
@@ -357,14 +323,14 @@ void PdfConverterImpl::Stop() {
   pdf_to_emf_converter_.reset();
 }
 
-void PdfConverterImpl::OnFailed(const std::string& error_message) {
+void PdfConverterImpl::OnFailed(std::string_view error_message) {
   DCHECK_CURRENTLY_ON(BrowserThread::UI);
   LOG(ERROR) << "Failed to convert PDF: " << error_message;
   base::WeakPtr<PdfConverterImpl> weak_this = weak_ptr_factory_.GetWeakPtr();
-  if (!start_callback_.is_null()) {
+  if (start_callback_) {
     std::move(start_callback_).Run(/*page_count=*/0);
     if (!weak_this)
-      return;  // Protect against the |start_callback_| deleting |this|.
+      return;  // Protect against the `start_callback_` deleting `this`.
   }
 
   while (!get_page_callbacks_.empty()) {
@@ -396,10 +362,6 @@ void PdfConverterImpl::RecordConversionMetrics() {
     case PdfRenderSettings::Mode::TEXTONLY:
       // Intentionally not logged.
       return;
-    case PdfRenderSettings::Mode::GDI_TEXT:
-      UMA_HISTOGRAM_MEMORY_KB("Printing.ConversionSize.EmfWithGdiText",
-                              average_page_size_in_kb);
-      return;
     case PdfRenderSettings::Mode::POSTSCRIPT_LEVEL2:
       UMA_HISTOGRAM_MEMORY_KB("Printing.ConversionSize.PostScript2",
                               average_page_size_in_kb);
@@ -408,8 +370,15 @@ void PdfConverterImpl::RecordConversionMetrics() {
       UMA_HISTOGRAM_MEMORY_KB("Printing.ConversionSize.PostScript3",
                               average_page_size_in_kb);
       return;
-    default:
-      NOTREACHED();
+    case PdfRenderSettings::Mode::EMF_WITH_REDUCED_RASTERIZATION:
+      UMA_HISTOGRAM_MEMORY_KB(
+          "Printing.ConversionSize.EmfWithReducedRasterization",
+          average_page_size_in_kb);
+      return;
+    case PdfRenderSettings::Mode::POSTSCRIPT_LEVEL3_WITH_TYPE42_FONTS:
+      UMA_HISTOGRAM_MEMORY_KB(
+          "Printing.ConversionSize.PostScript3WithType42Fonts",
+          average_page_size_in_kb);
       return;
   }
 }
@@ -422,9 +391,11 @@ PdfConverter::~PdfConverter() = default;
 std::unique_ptr<PdfConverter> PdfConverter::StartPdfConverter(
     scoped_refptr<base::RefCountedMemory> data,
     const PdfRenderSettings& conversion_settings,
+    const absl::optional<bool>& use_skia,
+    const GURL& url,
     StartCallback start_callback) {
-  return std::make_unique<PdfConverterImpl>(data, conversion_settings,
-                                            std::move(start_callback));
+  return std::make_unique<PdfConverterImpl>(data, conversion_settings, use_skia,
+                                            url, std::move(start_callback));
 }
 
 ScopedSimulateFailureCreatingTempFileForTests::
