@@ -1,52 +1,53 @@
-// Copyright (c) 2012 The Chromium Authors. All rights reserved.
+// Copyright 2012 The Chromium Authors
 // Use of this source code is governed by a BSD-style license that can be
 // found in the LICENSE file.
 
 #include "chrome/browser/printing/print_preview_dialog_controller.h"
 
-#include <stddef.h>
-
 #include <algorithm>
+#include <memory>
 #include <string>
 #include <vector>
 
 #include "base/auto_reset.h"
-#include "base/macros.h"
-#include "base/path_service.h"
-#include "base/strings/utf_string_conversions.h"
+#include "base/containers/contains.h"
+#include "base/memory/weak_ptr.h"
+#include "build/branding_buildflags.h"
 #include "build/build_config.h"
+#include "build/chromeos_buildflags.h"
 #include "chrome/browser/browser_process.h"
-#include "chrome/browser/extensions/chrome_extension_web_contents_observer.h"
 #include "chrome/browser/printing/print_view_manager.h"
+#include "chrome/browser/printing/print_view_manager_base.h"
 #include "chrome/browser/task_manager/web_contents_tags.h"
 #include "chrome/browser/ui/browser.h"
 #include "chrome/browser/ui/browser_finder.h"
 #include "chrome/browser/ui/browser_window.h"
-#include "chrome/browser/ui/webui/chrome_web_contents_handler.h"
 #include "chrome/browser/ui/webui/constrained_web_dialog_ui.h"
 #include "chrome/browser/ui/webui/print_preview/print_preview_ui.h"
-#include "chrome/common/chrome_features.h"
-#include "chrome/common/chrome_paths.h"
-#include "chrome/common/url_constants.h"
+#include "chrome/common/webui_url_constants.h"
 #include "chrome/grit/generated_resources.h"
 #include "components/guest_view/browser/guest_view_base.h"
 #include "components/web_modal/web_contents_modal_dialog_host.h"
 #include "content/public/browser/host_zoom_map.h"
-#include "content/public/browser/navigation_controller.h"
-#include "content/public/browser/navigation_details.h"
 #include "content/public/browser/navigation_entry.h"
+#include "content/public/browser/navigation_handle.h"
 #include "content/public/browser/render_frame_host.h"
 #include "content/public/browser/web_contents.h"
-#include "content/public/browser/web_contents_delegate.h"
-#include "content/public/browser/web_contents_observer.h"
+#include "content/public/browser/web_ui.h"
+#include "content/public/common/url_constants.h"
 #include "ui/base/l10n/l10n_util.h"
 #include "ui/web_dialogs/web_dialog_delegate.h"
 
-#if defined(OS_WIN) && defined(GOOGLE_CHROME_BUILD)
+#if BUILDFLAG(IS_WIN) && BUILDFLAG(GOOGLE_CHROME_BRANDING)
 #include "chrome/browser/win/conflicts/module_database.h"
 #endif
 
+#if BUILDFLAG(IS_CHROMEOS_ASH)
+#include "chrome/browser/ash/arc/print_spooler/print_session_impl.h"
+#endif
+
 using content::NavigationController;
+using content::NavigationHandle;
 using content::WebContents;
 using content::WebUIMessageHandler;
 
@@ -56,20 +57,34 @@ namespace {
 
 PrintPreviewUI* GetPrintPreviewUIForDialog(WebContents* dialog) {
   content::WebUI* web_ui = dialog->GetWebUI();
-  return web_ui ? static_cast<PrintPreviewUI*>(web_ui->GetController())
-                : nullptr;
+  return web_ui ? web_ui->GetController()->GetAs<PrintPreviewUI>() : nullptr;
 }
 
+#if BUILDFLAG(IS_CHROMEOS_ASH)
+void CloseArcPrintSession(WebContents* initiator) {
+  WebContents* outermost_web_contents =
+      guest_view::GuestViewBase::GetTopLevelWebContents(initiator);
+  auto* arc_print_session =
+      arc::PrintSessionImpl::FromWebContents(outermost_web_contents);
+  if (arc_print_session)
+    arc_print_session->OnPrintPreviewClosed();
+}
+#endif
+
 // A ui::WebDialogDelegate that specifies the print preview dialog appearance.
-class PrintPreviewDialogDelegate : public ui::WebDialogDelegate,
-                                   public content::WebContentsObserver {
+class PrintPreviewDialogDelegate : public ui::WebDialogDelegate {
  public:
   explicit PrintPreviewDialogDelegate(WebContents* initiator);
+
+  PrintPreviewDialogDelegate(const PrintPreviewDialogDelegate&) = delete;
+  PrintPreviewDialogDelegate& operator=(const PrintPreviewDialogDelegate&) =
+      delete;
+
   ~PrintPreviewDialogDelegate() override;
 
   ui::ModalType GetDialogModalType() const override;
-  base::string16 GetDialogTitle() const override;
-  base::string16 GetAccessibleDialogTitle() const override;
+  std::u16string GetDialogTitle() const override;
+  std::u16string GetAccessibleDialogTitle() const override;
   GURL GetDialogContentURL() const override;
   void GetWebUIMessageHandlers(
       std::vector<WebUIMessageHandler*>* handlers) const override;
@@ -81,15 +96,14 @@ class PrintPreviewDialogDelegate : public ui::WebDialogDelegate,
   bool ShouldShowDialogTitle() const override;
 
  private:
-  WebContents* initiator() const { return web_contents(); }
+  WebContents* initiator() const { return web_contents_.get(); }
 
+  base::WeakPtr<content::WebContents> web_contents_;
   bool on_dialog_closed_called_ = false;
-
-  DISALLOW_COPY_AND_ASSIGN(PrintPreviewDialogDelegate);
 };
 
 PrintPreviewDialogDelegate::PrintPreviewDialogDelegate(WebContents* initiator)
-    : content::WebContentsObserver(initiator) {}
+    : web_contents_(initiator->GetWeakPtr()) {}
 
 PrintPreviewDialogDelegate::~PrintPreviewDialogDelegate() = default;
 
@@ -99,12 +113,12 @@ ui::ModalType PrintPreviewDialogDelegate::GetDialogModalType() const {
   return ui::MODAL_TYPE_WINDOW;
 }
 
-base::string16 PrintPreviewDialogDelegate::GetDialogTitle() const {
+std::u16string PrintPreviewDialogDelegate::GetDialogTitle() const {
   // Only used on Windows? UI folks prefer no title.
-  return base::string16();
+  return std::u16string();
 }
 
-base::string16 PrintPreviewDialogDelegate::GetAccessibleDialogTitle() const {
+std::u16string PrintPreviewDialogDelegate::GetAccessibleDialogTitle() const {
   return l10n_util::GetStringUTF16(IDS_PRINT_PREVIEW_TITLE);
 }
 
@@ -140,16 +154,11 @@ void PrintPreviewDialogDelegate::GetDialogSize(gfx::Size* size) const {
   size->Enlarge(-2 * kBorder, -kBorder);
 
   static const gfx::Size kMaxDialogSize(1000, 660);
-  bool should_limit_dialog_size =
-      base::FeatureList::IsEnabled(::features::kNewPrintPreviewLayout);
-#if defined(OS_MACOSX)
-  // Limit the maximum size on MacOS X.
-  // http://crbug.com/105815
-  should_limit_dialog_size = true;
-#endif
-  if (should_limit_dialog_size) {
-    size->SetToMin(kMaxDialogSize);
-  }
+  int max_width = std::max(size->width() * 7 / 10, kMaxDialogSize.width());
+  int max_height =
+      std::max(max_width * kMaxDialogSize.height() / kMaxDialogSize.width(),
+               kMaxDialogSize.height());
+  size->SetToMin(gfx::Size(max_width, max_height));
 }
 
 std::string PrintPreviewDialogDelegate::GetDialogArgs() const {
@@ -183,54 +192,10 @@ bool PrintPreviewDialogDelegate::ShouldShowDialogTitle() const {
 
 }  // namespace
 
-// PrintPreviewDialogController::WebContentsObserver ---------------------------
-
-class PrintPreviewDialogController::WebContentsObserver
-    : public content::WebContentsObserver {
- public:
-  WebContentsObserver(PrintPreviewDialogController* controller,
-                      content::WebContents* web_contents);
-  ~WebContentsObserver() override;
-
-  // content::WebContentsObserver:
-  void RenderProcessGone(base::TerminationStatus status) override;
-  void NavigationEntryCommitted(
-      const content::LoadCommittedDetails& load_details) override;
-  void WebContentsDestroyed() override;
-
- private:
-  PrintPreviewDialogController* const controller_;
-
-  DISALLOW_COPY_AND_ASSIGN(WebContentsObserver);
-};
-
-PrintPreviewDialogController::WebContentsObserver::WebContentsObserver(
-    PrintPreviewDialogController* controller,
-    content::WebContents* web_contents)
-    : content::WebContentsObserver(web_contents), controller_(controller) {}
-
-PrintPreviewDialogController::WebContentsObserver::~WebContentsObserver() =
-    default;
-
-void PrintPreviewDialogController::WebContentsObserver::RenderProcessGone(
-    base::TerminationStatus status) {
-  controller_->OnRendererProcessClosed(
-      web_contents()->GetMainFrame()->GetProcess());
-}
-
-void PrintPreviewDialogController::WebContentsObserver::
-    NavigationEntryCommitted(
-        const content::LoadCommittedDetails& load_details) {
-  controller_->OnNavEntryCommitted(web_contents(), load_details);
-}
-
-void PrintPreviewDialogController::WebContentsObserver::WebContentsDestroyed() {
-  controller_->OnWebContentsDestroyed(web_contents());
-}
-
 // PrintPreviewDialogController ------------------------------------------------
 
-PrintPreviewDialogController::PrintPreviewDialogController() = default;
+PrintPreviewDialogController::PrintPreviewDialogController()
+    : web_contents_collection_(this) {}
 
 // static
 PrintPreviewDialogController* PrintPreviewDialogController::GetInstance() {
@@ -239,55 +204,52 @@ PrintPreviewDialogController* PrintPreviewDialogController::GetInstance() {
   return g_browser_process->print_preview_dialog_controller();
 }
 
-// static
 void PrintPreviewDialogController::PrintPreview(WebContents* initiator) {
-#if defined(OS_WIN) && defined(GOOGLE_CHROME_BUILD)
-  ModuleDatabase::GetInstance()->DisableThirdPartyBlocking();
+#if BUILDFLAG(IS_WIN) && BUILDFLAG(GOOGLE_CHROME_BRANDING)
+  PrintViewManagerBase::DisableThirdPartyBlocking();
 #endif
 
-  if (initiator->ShowingInterstitialPage() || initiator->IsCrashed())
+  if (initiator->IsCrashed()) {
     return;
-
-  PrintPreviewDialogController* dialog_controller = GetInstance();
-  if (!dialog_controller)
-    return;
-  if (!dialog_controller->GetOrCreatePreviewDialog(initiator)) {
-    PrintViewManager* print_view_manager =
-        PrintViewManager::FromWebContents(initiator);
-    if (print_view_manager)
-      print_view_manager->PrintPreviewDone();
   }
+
+  if (!GetOrCreatePreviewDialog(initiator)) {
+    auto* print_view_manager = PrintViewManager::FromWebContents(initiator);
+    if (print_view_manager) {
+      print_view_manager->PrintPreviewDone();
+    }
+  }
+}
+
+WebContents* PrintPreviewDialogController::GetOrCreatePreviewDialogForTesting(
+    WebContents* initiator) {
+  return GetOrCreatePreviewDialog(initiator);
 }
 
 WebContents* PrintPreviewDialogController::GetOrCreatePreviewDialog(
     WebContents* initiator) {
   DCHECK(initiator);
 
-  // Get the print preview dialog for |initiator|.
+  // Get the print preview dialog for `initiator`.
   WebContents* preview_dialog = GetPrintPreviewForContents(initiator);
-  if (!preview_dialog)
-    return CreatePrintPreviewDialog(initiator);
-
-  // Show the initiator holding the existing preview dialog.
-  initiator->GetDelegate()->ActivateContents(initiator);
-  return preview_dialog;
+  if (preview_dialog) {
+    return preview_dialog;
+  }
+  return CreatePrintPreviewDialog(initiator);
 }
 
 WebContents* PrintPreviewDialogController::GetPrintPreviewForContents(
     WebContents* contents) const {
-  // |preview_dialog_map_| is keyed by the preview dialog, so if find()
-  // succeeds, then |contents| is the preview dialog.
-  auto it = preview_dialog_map_.find(contents);
-  if (it != preview_dialog_map_.end())
+  // `preview_dialog_map_` is keyed by the preview dialog, so if
+  // base::Contains() succeeds, then `contents` is the preview dialog.
+  if (base::Contains(preview_dialog_map_, contents))
     return contents;
 
-  for (it = preview_dialog_map_.begin();
-       it != preview_dialog_map_.end();
-       ++it) {
-    // If |contents| is an initiator.
-    if (contents == it->second) {
+  for (const auto& it : preview_dialog_map_) {
+    // If `contents` is an initiator.
+    if (contents == it.second) {
       // Return the associated preview dialog.
-      return it->first;
+      return it.first;
     }
   }
   return nullptr;
@@ -296,7 +258,7 @@ WebContents* PrintPreviewDialogController::GetPrintPreviewForContents(
 WebContents* PrintPreviewDialogController::GetInitiator(
     WebContents* preview_dialog) {
   auto it = preview_dialog_map_.find(preview_dialog);
-  return (it != preview_dialog_map_.end()) ? it->second : nullptr;
+  return it != preview_dialog_map_.end() ? it->second : nullptr;
 }
 
 void PrintPreviewDialogController::ForEachPreviewDialog(
@@ -307,8 +269,14 @@ void PrintPreviewDialogController::ForEachPreviewDialog(
 
 // static
 bool PrintPreviewDialogController::IsPrintPreviewURL(const GURL& url) {
-  return (url.SchemeIs(content::kChromeUIScheme) &&
-          url.host_piece() == chrome::kChromeUIPrintHost);
+  return url.SchemeIs(content::kChromeUIScheme) &&
+         url.host_piece() == chrome::kChromeUIPrintHost;
+}
+
+// static
+bool PrintPreviewDialogController::IsPrintPreviewContentURL(const GURL& url) {
+  return url.SchemeIs(content::kChromeUIUntrustedScheme) &&
+         url.host_piece() == chrome::kChromeUIPrintHost;
 }
 
 void PrintPreviewDialogController::EraseInitiatorInfo(
@@ -317,24 +285,28 @@ void PrintPreviewDialogController::EraseInitiatorInfo(
   if (it == preview_dialog_map_.end())
     return;
 
-  RemoveObserver(it->second);
-  preview_dialog_map_[preview_dialog] = nullptr;
+  web_contents_collection_.StopObserving(it->second);
+  it->second = nullptr;
 }
 
 PrintPreviewDialogController::~PrintPreviewDialogController() = default;
 
-void PrintPreviewDialogController::OnRendererProcessClosed(
-    content::RenderProcessHost* rph) {
+void PrintPreviewDialogController::RenderProcessGone(
+    content::WebContents* web_contents,
+    base::TerminationStatus status) {
+  content::RenderProcessHost* rph =
+      web_contents->GetPrimaryMainFrame()->GetProcess();
+
   // Store contents in a vector and deal with them after iterating through
-  // |preview_dialog_map_| because RemoveFoo() can change |preview_dialog_map_|.
+  // `preview_dialog_map_` because RemoveFoo() can change `preview_dialog_map_`.
   std::vector<WebContents*> closed_initiators;
   std::vector<WebContents*> closed_preview_dialogs;
   for (auto& it : preview_dialog_map_) {
     WebContents* preview_dialog = it.first;
     WebContents* initiator = it.second;
-    if (preview_dialog->GetMainFrame()->GetProcess() == rph)
+    if (preview_dialog->GetPrimaryMainFrame()->GetProcess() == rph)
       closed_preview_dialogs.push_back(preview_dialog);
-    else if (initiator && initiator->GetMainFrame()->GetProcess() == rph)
+    else if (initiator && initiator->GetPrimaryMainFrame()->GetProcess() == rph)
       closed_initiators.push_back(initiator);
   }
 
@@ -349,8 +321,7 @@ void PrintPreviewDialogController::OnRendererProcessClosed(
     RemoveInitiator(initiator);
 }
 
-void PrintPreviewDialogController::OnWebContentsDestroyed(
-    WebContents* contents) {
+void PrintPreviewDialogController::WebContentsDestroyed(WebContents* contents) {
   WebContents* preview_dialog = GetPrintPreviewForContents(contents);
   if (!preview_dialog) {
     NOTREACHED();
@@ -363,9 +334,14 @@ void PrintPreviewDialogController::OnWebContentsDestroyed(
     RemoveInitiator(contents);
 }
 
-void PrintPreviewDialogController::OnNavEntryCommitted(
+void PrintPreviewDialogController::DidFinishNavigation(
     WebContents* contents,
-    const content::LoadCommittedDetails& details) {
+    NavigationHandle* navigation_handle) {
+  if (!navigation_handle->IsInPrimaryMainFrame() ||
+      !navigation_handle->HasCommitted()) {
+    return;
+  }
+
   WebContents* preview_dialog = GetPrintPreviewForContents(contents);
   if (!preview_dialog) {
     NOTREACHED();
@@ -373,21 +349,21 @@ void PrintPreviewDialogController::OnNavEntryCommitted(
   }
 
   if (contents != preview_dialog)
-    OnInitiatorNavigated(contents, details);
+    OnInitiatorNavigated(contents, navigation_handle);
   else
-    OnPreviewDialogNavigated(contents, details);
+    OnPreviewDialogNavigated(contents, navigation_handle);
 }
 
 void PrintPreviewDialogController::OnInitiatorNavigated(
     WebContents* initiator,
-    const content::LoadCommittedDetails& details) {
-  if (details.type == content::NAVIGATION_TYPE_EXISTING_PAGE) {
+    NavigationHandle* navigation_handle) {
+  if (navigation_handle->IsSameDocument()) {
     static const ui::PageTransition kTransitions[] = {
         ui::PageTransitionFromInt(ui::PAGE_TRANSITION_TYPED |
                                   ui::PAGE_TRANSITION_FROM_ADDRESS_BAR),
-        ui::PAGE_TRANSITION_LINK,
-    };
-    ui::PageTransition type = details.entry->GetTransitionType();
+        ui::PAGE_TRANSITION_LINK, ui::PAGE_TRANSITION_AUTO_BOOKMARK};
+    ui::PageTransition type =
+        initiator->GetController().GetLastCommittedEntry()->GetTransitionType();
     for (ui::PageTransition transition : kTransitions) {
       if (ui::PageTransitionTypeIncludingQualifiersIs(type, transition))
         return;
@@ -399,27 +375,18 @@ void PrintPreviewDialogController::OnInitiatorNavigated(
 
 void PrintPreviewDialogController::OnPreviewDialogNavigated(
     WebContents* preview_dialog,
-    const content::LoadCommittedDetails& details) {
-  ui::PageTransition type = details.entry->GetTransitionType();
+    NavigationHandle* navigation_handle) {
+  ui::PageTransition type = preview_dialog->GetController()
+                                .GetLastCommittedEntry()
+                                ->GetTransitionType();
 
-  // New |preview_dialog| is created. Don't update/erase map entry.
-  if (waiting_for_new_preview_page_ &&
+  // New `preview_dialog` is created. Don't update/erase map entry.
+  if (navigation_handle->GetPreviousPrimaryMainFrameURL().is_empty() &&
+      IsPrintPreviewURL(navigation_handle->GetURL()) &&
       ui::PageTransitionCoreTypeIs(type, ui::PAGE_TRANSITION_AUTO_TOPLEVEL) &&
-      details.type == content::NAVIGATION_TYPE_NEW_PAGE) {
-    waiting_for_new_preview_page_ = false;
+      !navigation_handle->IsSameDocument()) {
     SaveInitiatorTitle(preview_dialog);
-    return;
   }
-
-  // Cloud print sign-in causes a reload.
-  if (!waiting_for_new_preview_page_ &&
-      ui::PageTransitionCoreTypeIs(type, ui::PAGE_TRANSITION_RELOAD) &&
-      details.type == content::NAVIGATION_TYPE_EXISTING_PAGE &&
-      IsPrintPreviewURL(details.previous_url)) {
-    return;
-  }
-
-  NOTREACHED();
 }
 
 WebContents* PrintPreviewDialogController::CreatePrintPreviewDialog(
@@ -440,18 +407,15 @@ WebContents* PrintPreviewDialogController::CreatePrintPreviewDialog(
   content::HostZoomMap::Get(preview_dialog->GetSiteInstance())
       ->SetZoomLevelForHostAndScheme(print_url.scheme(), print_url.host(), 0);
   PrintViewManager::CreateForWebContents(preview_dialog);
-  extensions::ChromeExtensionWebContentsObserver::CreateForWebContents(
-      preview_dialog);
 
   // Add an entry to the map.
   preview_dialog_map_[preview_dialog] = initiator;
-  waiting_for_new_preview_page_ = true;
 
   // Make the print preview WebContents show up in the task manager.
   task_manager::WebContentsTags::CreateForPrintingContents(preview_dialog);
 
-  AddObserver(initiator);
-  AddObserver(preview_dialog);
+  web_contents_collection_.StartObserving(initiator);
+  web_contents_collection_.StartObserving(preview_dialog);
 
   return preview_dialog;
 }
@@ -470,18 +434,6 @@ void PrintPreviewDialogController::SaveInitiatorTitle(
       PrintViewManager::FromWebContents(initiator)->RenderSourceName());
 }
 
-void PrintPreviewDialogController::AddObserver(WebContents* contents) {
-  auto emplace_result = web_contents_observers_.emplace(
-      std::piecewise_construct, std::forward_as_tuple(contents),
-      std::forward_as_tuple(
-          std::make_unique<WebContentsObserver>(this, contents)));
-  DCHECK(emplace_result.second);
-}
-
-void PrintPreviewDialogController::RemoveObserver(WebContents* contents) {
-  web_contents_observers_.erase(contents);
-}
-
 void PrintPreviewDialogController::RemoveInitiator(
     WebContents* initiator) {
   WebContents* preview_dialog = GetPrintPreviewForContents(initiator);
@@ -490,10 +442,14 @@ void PrintPreviewDialogController::RemoveInitiator(
   // and reaches RemovePreviewDialog(), it does not attempt to also remove the
   // initiator's observers.
   preview_dialog_map_[preview_dialog] = nullptr;
-  RemoveObserver(initiator);
+  web_contents_collection_.StopObserving(initiator);
 
   PrintViewManager::FromWebContents(initiator)->PrintPreviewDone();
 
+#if BUILDFLAG(IS_CHROMEOS_ASH)
+  CloseArcPrintSession(initiator);
+#endif
+
   // Initiator is closed. Close the print preview dialog too.
   auto* print_preview_ui = GetPrintPreviewUIForDialog(preview_dialog);
   if (print_preview_ui)
@@ -505,12 +461,16 @@ void PrintPreviewDialogController::RemovePreviewDialog(
   // Remove the initiator's observers before erasing the mapping.
   WebContents* initiator = GetInitiator(preview_dialog);
   if (initiator) {
-    RemoveObserver(initiator);
+    web_contents_collection_.StopObserving(initiator);
     PrintViewManager::FromWebContents(initiator)->PrintPreviewDone();
+
+#if BUILDFLAG(IS_CHROMEOS_ASH)
+    CloseArcPrintSession(initiator);
+#endif
   }
 
   preview_dialog_map_.erase(preview_dialog);
-  RemoveObserver(preview_dialog);
+  web_contents_collection_.StopObserving(preview_dialog);
 }
 
 }  // namespace printing
