@@ -1,4 +1,4 @@
-// Copyright (c) 2012 The Chromium Authors. All rights reserved.
+// Copyright 2012 The Chromium Authors
 // Use of this source code is governed by a BSD-style license that can be
 // found in the LICENSE file.
 
@@ -8,18 +8,17 @@
 
 #include <utility>
 
-#include "base/bind.h"
+#include "base/check_op.h"
 #include "base/files/file_util.h"
+#include "base/functional/bind.h"
 #include "base/lazy_instance.h"
 #include "base/location.h"
-#include "base/logging.h"
+#include "base/notreached.h"
+#include "base/observer_list.h"
 #include "base/path_service.h"
-#include "base/single_thread_task_runner.h"
-#include "base/task/post_task.h"
+#include "base/task/single_thread_task_runner.h"
 #include "base/task/thread_pool.h"
-#include "base/task_runner_util.h"
 #include "base/threading/scoped_blocking_call.h"
-#include "base/threading/thread_task_runner_handle.h"
 #include "build/build_config.h"
 #include "chrome/browser/spellchecker/spellcheck_service.h"
 #ifndef TOOLKIT_QT
@@ -37,11 +36,12 @@
 #include "net/traffic_annotation/network_traffic_annotation.h"
 #include "services/network/public/cpp/resource_request.h"
 #include "services/network/public/cpp/simple_url_loader.h"
+#include "services/network/public/mojom/url_response_head.mojom.h"
 #include "url/gurl.h"
 
-#if !defined(OS_ANDROID)
+#if !BUILDFLAG(IS_ANDROID)
 #include "base/files/memory_mapped_file.h"
-#include "third_party/hunspell/google/bdict.h"
+#include "third_party/hunspell/google/bdict.h"  // nogncheck crbug.com/1125897
 #endif
 
 using content::BrowserThread;
@@ -65,23 +65,20 @@ bool SaveDictionaryData(std::unique_ptr<std::string> data,
   base::ScopedBlockingCall scoped_blocking_call(FROM_HERE,
                                                 base::BlockingType::MAY_BLOCK);
 
-  size_t bytes_written =
-      base::WriteFile(path, data->data(), data->length());
-  if (bytes_written != data->length()) {
+  if (!base::WriteFile(path, *data)) {
     bool success = false;
-#if defined(OS_WIN)
+#if BUILDFLAG(IS_WIN)
     base::FilePath dict_dir;
     base::PathService::Get(base::DIR_USER_DATA, &dict_dir);
     base::FilePath fallback_file_path =
         dict_dir.Append(path.BaseName());
-    bytes_written =
-        base::WriteFile(fallback_file_path, data->data(), data->length());
-    if (bytes_written == data->length())
+    if (base::WriteFile(fallback_file_path, *data)) {
       success = true;
+    }
 #endif
 
     if (!success) {
-      base::DeleteFile(path, false);
+      base::DeleteFile(path);
       return false;
     }
   }
@@ -91,48 +88,54 @@ bool SaveDictionaryData(std::unique_ptr<std::string> data,
 
 }  // namespace
 
-SpellcheckHunspellDictionary::DictionaryFile::DictionaryFile() {
-}
+SpellcheckHunspellDictionary::DictionaryFile::DictionaryFile(
+    base::TaskRunner* task_runner) : task_runner_(task_runner) {}
 
 SpellcheckHunspellDictionary::DictionaryFile::~DictionaryFile() {
+  if (file.IsValid()) {
+    task_runner_->PostTask(FROM_HERE,
+                           base::BindOnce(&CloseDictionary, std::move(file)));
+  }
 }
 
 SpellcheckHunspellDictionary::DictionaryFile::DictionaryFile(
     DictionaryFile&& other)
-    : path(other.path), file(std::move(other.file)) {}
+    : path(other.path),
+      file(std::move(other.file)),
+      task_runner_(std::move(other.task_runner_)) {}
 
 SpellcheckHunspellDictionary::DictionaryFile&
 SpellcheckHunspellDictionary::DictionaryFile::operator=(
     DictionaryFile&& other) {
   path = other.path;
   file = std::move(other.file);
+  task_runner_ = std::move(other.task_runner_);
   return *this;
 }
 
 SpellcheckHunspellDictionary::SpellcheckHunspellDictionary(
     const std::string& language,
+    const std::string& platform_spellcheck_language,
     content::BrowserContext* browser_context,
     SpellcheckService* spellcheck_service)
     : task_runner_(
           base::ThreadPool::CreateSequencedTaskRunner({base::MayBlock()})),
       language_(language),
+      platform_spellcheck_language_(platform_spellcheck_language),
       use_browser_spellchecker_(false),
       browser_context_(browser_context),
       spellcheck_service_(spellcheck_service),
-      download_status_(DOWNLOAD_NONE) {}
+      download_status_(DOWNLOAD_NONE),
+      dictionary_file_(task_runner_.get()) {}
 
 SpellcheckHunspellDictionary::~SpellcheckHunspellDictionary() {
-  if (dictionary_file_.file.IsValid()) {
-    task_runner_->PostTask(
-        FROM_HERE,
-        base::BindOnce(&CloseDictionary, std::move(dictionary_file_.file)));
-  }
-
 #if BUILDFLAG(USE_BROWSER_SPELLCHECKER)
   // Disable the language from platform spellchecker.
-  if (spellcheck::UseBrowserSpellChecker())
+  if (spellcheck::UseBrowserSpellChecker() && HasPlatformSupport()) {
     spellcheck_platform::DisableLanguage(
-        spellcheck_service_->platform_spell_checker(), language_);
+        spellcheck_service_->platform_spell_checker(),
+        GetPlatformSpellcheckLanguage());
+  }
 #endif  // BUILDFLAG(USE_BROWSER_SPELLCHECKER)
 }
 
@@ -141,9 +144,10 @@ void SpellcheckHunspellDictionary::Load() {
 
 #if BUILDFLAG(USE_BROWSER_SPELLCHECKER)
   if (spellcheck::UseBrowserSpellChecker() &&
-      spellcheck_platform::SpellCheckerAvailable()) {
+      spellcheck_platform::SpellCheckerAvailable() && HasPlatformSupport()) {
     spellcheck_platform::PlatformSupportsLanguage(
-        spellcheck_service_->platform_spell_checker(), language_,
+        spellcheck_service_->platform_spell_checker(),
+        GetPlatformSpellcheckLanguage(),
         base::BindOnce(
             &SpellcheckHunspellDictionary::PlatformSupportsLanguageComplete,
             weak_ptr_factory_.GetWeakPtr()));
@@ -178,6 +182,21 @@ const std::string& SpellcheckHunspellDictionary::GetLanguage() const {
   return language_;
 }
 
+const std::string& SpellcheckHunspellDictionary::GetPlatformSpellcheckLanguage()
+    const {
+#if BUILDFLAG(IS_WIN) && BUILDFLAG(USE_BROWSER_SPELLCHECKER)
+  // Currently the platform spellcheck language is only distinguished for
+  // Windows.
+  return platform_spellcheck_language_;
+#else
+  return language_;
+#endif  // BUILDFLAG(IS_WIN) && BUILDFLAG(USE_BROWSER_SPELLCHECKER)
+}
+
+bool SpellcheckHunspellDictionary::HasPlatformSupport() const {
+  return !GetPlatformSpellcheckLanguage().empty();
+}
+
 bool SpellcheckHunspellDictionary::IsUsingPlatformChecker() const {
   return use_browser_spellchecker_;
 }
@@ -225,10 +244,9 @@ void SpellcheckHunspellDictionary::OnSimpleLoaderComplete(
     return;
   }
 
-#if !defined(OS_ANDROID)
+#if !BUILDFLAG(IS_ANDROID)
   // To prevent corrupted dictionary data from causing a renderer crash, scan
   // the dictionary data and verify it is sane before save it to a file.
-  // TODO(rlp): Adding metrics to RecordDictionaryCorruptionStats
   if (!hunspell::BDict::Verify(data->data(), data->size())) {
     // Let PostTaskAndReply caller send to InformListenersOfInitialization
     // through SaveDictionaryDataComplete().
@@ -237,8 +255,8 @@ void SpellcheckHunspellDictionary::OnSimpleLoaderComplete(
   }
 #endif
 
-  base::PostTaskAndReplyWithResult(
-      task_runner_.get(), FROM_HERE,
+  task_runner_->PostTaskAndReplyWithResult(
+      FROM_HERE,
       base::BindOnce(&SaveDictionaryData, std::move(data),
                      dictionary_file_.path),
       base::BindOnce(&SpellcheckHunspellDictionary::SaveDictionaryDataComplete,
@@ -300,7 +318,7 @@ void SpellcheckHunspellDictionary::DownloadDictionary(GURL url) {
   simple_loader_ = network::SimpleURLLoader::Create(std::move(resource_request),
                                                     traffic_annotation);
   network::mojom::URLLoaderFactory* loader_factory =
-      content::BrowserContext::GetDefaultStoragePartition(browser_context_)
+      browser_context_->GetDefaultStoragePartition()
           ->GetURLLoaderFactoryForBrowserProcess()
           .get();
   simple_loader_->DownloadToStringOfUnboundedSizeUntilCrashAndDie(
@@ -309,13 +327,14 @@ void SpellcheckHunspellDictionary::DownloadDictionary(GURL url) {
                      base::Unretained(this)));
 
   // Attempt downloading the dictionary only once.
-  browser_context_ = NULL;
+  browser_context_ = nullptr;
 }
 
-#if !defined(OS_ANDROID)
+#if !BUILDFLAG(IS_ANDROID)
 // static
 SpellcheckHunspellDictionary::DictionaryFile
-SpellcheckHunspellDictionary::OpenDictionaryFile(const base::FilePath& path) {
+SpellcheckHunspellDictionary::OpenDictionaryFile(base::TaskRunner* task_runner,
+                                                 const base::FilePath& path) {
   base::ScopedBlockingCall scoped_blocking_call(FROM_HERE,
                                                 base::BlockingType::MAY_BLOCK);
 
@@ -326,9 +345,9 @@ SpellcheckHunspellDictionary::OpenDictionaryFile(const base::FilePath& path) {
   // For systemwide installations on Windows, the default directory may not
   // have permissions for download. In that case, the alternate directory for
   // download is chrome::DIR_USER_DATA.
-  DictionaryFile dictionary;
+  DictionaryFile dictionary(task_runner);
 
-#if defined(OS_WIN)
+#if BUILDFLAG(IS_WIN)
   // Check if the dictionary exists in the fallback location. If so, use it
   // rather than downloading anew.
   base::FilePath user_dir;
@@ -340,7 +359,7 @@ SpellcheckHunspellDictionary::OpenDictionaryFile(const base::FilePath& path) {
     dictionary.path = path;
 #else
   dictionary.path = path;
-#endif  // defined(OS_WIN)
+#endif  // BUILDFLAG(IS_WIN)
 
   // Read the dictionary file and scan its data to check for corruption. The
   // scoping closes the memory-mapped file before it is opened or deleted.
@@ -360,7 +379,7 @@ SpellcheckHunspellDictionary::OpenDictionaryFile(const base::FilePath& path) {
                                base::File::FLAG_READ | base::File::FLAG_OPEN);
   } else {
 #ifndef TOOLKIT_QT
-    base::DeleteFile(dictionary.path, false);
+    base::DeleteFile(dictionary.path);
 #endif
   }
 
@@ -370,7 +389,7 @@ SpellcheckHunspellDictionary::OpenDictionaryFile(const base::FilePath& path) {
 // static
 SpellcheckHunspellDictionary::DictionaryFile
 SpellcheckHunspellDictionary::InitializeDictionaryLocation(
-    const std::string& language) {
+    base::TaskRunner* task_runner, const std::string& language) {
   base::ScopedBlockingCall scoped_blocking_call(FROM_HERE,
                                                 base::BlockingType::MAY_BLOCK);
 
@@ -385,7 +404,7 @@ SpellcheckHunspellDictionary::InitializeDictionaryLocation(
   base::FilePath dict_path =
       spellcheck::GetVersionedFileName(language, dict_dir);
 
-  return OpenDictionaryFile(dict_path);
+  return OpenDictionaryFile(task_runner, dict_path);
 }
 
 void SpellcheckHunspellDictionary::InitializeDictionaryLocationComplete(
@@ -399,7 +418,7 @@ void SpellcheckHunspellDictionary::InitializeDictionaryLocationComplete(
     // TODO(rouslan): Remove this test-only case.
     if (spellcheck_service_->SignalStatusEvent(
           SpellcheckService::BDICT_CORRUPTED)) {
-      browser_context_ = NULL;
+      browser_context_ = nullptr;
     }
 
     if (browser_context_) {
@@ -419,7 +438,7 @@ void SpellcheckHunspellDictionary::InitializeDictionaryLocationComplete(
       InformListenersOfInitialization();
 #endif
 }
-#endif  // !defined(OS_ANDROID)
+#endif  // !BUILDFLAG(IS_ANDROID)
 
 void SpellcheckHunspellDictionary::SaveDictionaryDataComplete(
     bool dictionary_saved) {
@@ -453,9 +472,10 @@ void SpellcheckHunspellDictionary::PlatformSupportsLanguageComplete(
 
   if (platform_supports_language) {
 #if BUILDFLAG(USE_BROWSER_SPELLCHECKER)
-    if (spellcheck::UseBrowserSpellChecker()) {
+    if (spellcheck::UseBrowserSpellChecker() && HasPlatformSupport()) {
       spellcheck_platform::SetLanguage(
-          spellcheck_service_->platform_spell_checker(), language_,
+          spellcheck_service_->platform_spell_checker(),
+          GetPlatformSpellcheckLanguage(),
           base::BindOnce(&SpellcheckHunspellDictionary::
                              SpellCheckPlatformSetLanguageComplete,
                          weak_ptr_factory_.GetWeakPtr()));
@@ -464,28 +484,18 @@ void SpellcheckHunspellDictionary::PlatformSupportsLanguageComplete(
 #endif  // BUILDFLAG(USE_BROWSER_SPELLCHECKER)
     NOTREACHED();
   } else {
-#if defined(OS_WIN) && BUILDFLAG(USE_BROWSER_SPELLCHECKER)
-    // The platform spellchecker doesn't support this language. Fall back to
-    // Hunspell, unless the hybrid spellchecker is disabled.
-    if (spellcheck::UseBrowserSpellChecker() &&
-        !spellcheck::UseWinHybridSpellChecker()) {
-      // Can't fall back to Hunspell, because the hybrid spellchecker is not
-      // enabled. We can't spellcheck this language, so there's no further
-      // processing to do.
-      return;
-    }
-#endif  // defined(OS_WIN) && BUILDFLAG(USE_BROWSER_SPELLCHECKER)
     // Either the platform spellchecker is unavailable / disabled, or it doesn't
     // support this language. In either case, we must use Hunspell for this
     // language, unless we are on Android, which doesn't support Hunspell.
-#if !defined(OS_ANDROID) && BUILDFLAG(USE_RENDERER_SPELLCHECKER)
-    base::PostTaskAndReplyWithResult(
-        task_runner_.get(), FROM_HERE,
-        base::BindOnce(&InitializeDictionaryLocation, language_),
+#if !BUILDFLAG(IS_ANDROID) && BUILDFLAG(USE_RENDERER_SPELLCHECKER)
+    task_runner_->PostTaskAndReplyWithResult(
+        FROM_HERE,
+        base::BindOnce(&InitializeDictionaryLocation,
+                       base::RetainedRef(task_runner_.get()), language_),
         base::BindOnce(
             &SpellcheckHunspellDictionary::InitializeDictionaryLocationComplete,
             weak_ptr_factory_.GetWeakPtr()));
-#endif  // !defined(OS_ANDROID) && BUILDFLAG(USE_RENDERER_SPELLCHECKER)
+#endif  // !BUILDFLAG(IS_ANDROID) && BUILDFLAG(USE_RENDERER_SPELLCHECKER)
   }
 }
 
@@ -498,7 +508,7 @@ void SpellcheckHunspellDictionary::SpellCheckPlatformSetLanguageComplete(
     return;
 
   use_browser_spellchecker_ = true;
-  base::ThreadTaskRunnerHandle::Get()->PostTask(
+  base::SingleThreadTaskRunner::GetCurrentDefault()->PostTask(
       FROM_HERE,
       base::BindOnce(
           &SpellcheckHunspellDictionary::InformListenersOfInitialization,
